"""Plots to monitor training."""

__author__ = 'Felix Kleinert, Lukas Leufen'
__date__ = '2019-12-11'

from typing import Union, Dict, List

import keras
import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from mlair.model_modules.keras_extensions import LearningRateDecay

# matplotlib.use('Agg')
history_object = Union[Dict, keras.callbacks.History]
lr_object = Union[Dict, LearningRateDecay]


class PlotModelHistory:
    """
    Plot history of all plot_metrics (default: loss) for a training event.

    For default plot_metric and val_plot_metric are plotted. If further metrics are provided (name must somehow include
    the word `<plot_metric>`), this additional information is added to the plot with an separate y-axis scale on the
    right side (shared for all additional metrics). The plot is saved locally. For a proper saving behaviour, the
    parameter filename must include the absolute path for the plot.
    """

    def __init__(self, filename: str, history: history_object, plot_metric: str = "loss", main_branch: bool = False):
        """
        Set attributes and create plot.

        :param filename: saving name of the plot to create (preferably absolute path if possible), the filename needs a
            format ending like .pdf or .png to work.
        :param history: the history object (or a dict with at least 'loss' and 'val_loss' as keys) to plot loss from
        :param plot_metric: the metric to plot (e.b. mean_squared_error, mse, mean_absolute_error, loss, default: loss)
        :param main_branch: switch between only looking for metrics that go with 'main' or for all occurrences (default:
            False -> look for losses from all branches, not only from main)
        """
        if isinstance(history, keras.callbacks.History):
            history = history.history
        self._data = pd.DataFrame.from_dict(history)
        self._plot_metric = self._get_plot_metric(history, plot_metric, main_branch)
        self._additional_columns = self._filter_columns(history)
        self._plot(filename)

    @staticmethod
    def _get_plot_metric(history, plot_metric, main_branch):
        if plot_metric.lower() == "mse":
            plot_metric = "mean_squared_error"
        elif plot_metric.lower() == "mae":
            plot_metric = "mean_absolute_error"
        available_keys = [k for k in history.keys() if
                          plot_metric in k and ("main" in k.lower() if main_branch else True)]
        available_keys.sort(key=len)
        return available_keys[0]

    def _filter_columns(self, history: Dict) -> List[str]:
        """
        Select only columns named like %<plot_metric>%.

        The default metrics '<plot_metric>' and 'val_<plot_metric>' are removed too.

        :param history: a dict with at least '<plot_metric>' and 'val_<plot_metric>' as keys (can be derived from keras
            History.history)

        :return: filtered columns including all plot_metric variations except <plot_metric> and val_<plot_metric>.
        """
        cols = list(filter(lambda x: self._plot_metric in x, history.keys()))
        # heuristic: there is always val_<plot_metric> and <plot_metric> available in cols, because this is generated by
        # the keras framework. If this metric isn't available the self._get_plot_metric() will fail before (but only
        # because it is executed before)
        cols.remove(f"val_{self._plot_metric}")
        cols.remove(self._plot_metric)
        return cols

    def _plot(self, filename: str) -> None:
        """
        Create plot.

        Plots <plot_metric> and val_<plot_metric> as default. If more plot_metrics are provided, they will be added with
        an additional yaxis on the right side. The plot is saved in filename.

        :param filename: name (including total path) of the plot to save.
        """
        ax = self._data[[self._plot_metric, f"val_{self._plot_metric}"]].plot(linewidth=0.7, logy=True)
        if len(self._additional_columns) > 0:
            self._data[self._additional_columns].plot(linewidth=0.7, secondary_y=True, ax=ax, logy=True)
        title = f"Model {self._plot_metric}: best = {self._data[[f'val_{self._plot_metric}']].min().values}"
        ax.set(xlabel="epoch", ylabel=self._plot_metric, title=title)
        ax.axhline(y=0, color="gray", linewidth=0.5)
        plt.tight_layout()
        plt.savefig(filename)
        plt.close("all")


class PlotModelLearningRate:
    """
    Plot the behaviour of the learning rate in dependence of the number of epochs.

    The plot is saved locally as pdf. For a proper saving behaviour, the parameter filename must include the absolute
    path for the plot.
    """

    def __init__(self, filename: str, lr_sc: lr_object):
        """
        Set attributes and create plot.

        :param filename: saving name of the plot to create (preferably absolute path if possible), the filename needs a
            format ending like .pdf or .png to work.
        :param lr_sc: the learning rate object (or a dict with `lr` as key) to plot from
        """
        if isinstance(lr_sc, LearningRateDecay):
            lr_sc = lr_sc.lr
        self._data = pd.DataFrame.from_dict(lr_sc)
        self._plot(filename)

    def _plot(self, filename: str) -> None:
        """
        Create plot.

        Plot the learning rate in dependence of epoch.

        :param filename: name (including total path) of the plot to save.
        """
        ax = self._data.plot(linewidth=0.7)
        ax.set(xlabel="epoch", ylabel="learning rate")
        plt.tight_layout()
        plt.savefig(filename)
        plt.close("all")
